/*
 * FILE:    arc.h
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#ifndef ARC_H
#define ARC_H

#include "connectable.h"

#include <QString>
#include <QPainterPath>
#include <QVector>
#include <QPoint>

class QPainter;

class Arc : public Drawable {
public:
  enum Type { Standard, Reset, Inhibitor, StoreAction };
  enum StoreActionE {
    ActionC = 0x0001,
    ActionR = 0x0002,
    ActionU = 0x0004,
    ActionD = 0x0008,
  };
  Q_DECLARE_FLAGS(StoreActions, StoreActionE)

  Arc();
  virtual ~Arc ();
  void draw (QPainter *p, DrawProps *drawProps); // Pre: p.isActive() && drawProps
  void drawPolylineWithArrows(QPainter *p, QPolygonF pointList, QPainterPath startArrow, QPainterPath endArrow);
  void drawArrow(QPainter *p, QPainterPath arrow, QPointF from, QPointF to);
  virtual void updateLabels(LabelPlacer *labelPlacer, DrawProps *drawProps); // Pre: labelPlacer
  virtual bool containsPos(QPoint pos) const;
  
  QRect getBoundingRect();
  virtual const QString getToolTip() const;
  void merge(Arc *other, bool swapTargets); // Pre: other && other->getType() == this->getType()
  
  inline Connectable *getSource() const;
  inline void setSource (Connectable *value);
  inline Connectable *getTarget() const;
  inline void setTarget (Connectable *value);
  
  inline int getProduction() const;
  inline void updateProduction(int add);
  inline int getConsumption() const;
  inline void updateConsumption(int add);

  inline const Type getType() const;
  inline void setType (Type value);
  inline const StoreActions getStoreAction() const;
  inline void setStoreAction(StoreActions value);
  
  const QVector<QPoint> getPoints();
  void setPoints(QVector<QPoint> value);
  void addPoint(QPoint value);
  
protected:
  QVector<QPointF> screenPointList() const;
  // Pre:  source && target
  // Post: ret.size() >= 2
  
  Connectable *source;
  Connectable *target;
  int consumption;
  int production;
  
  Type type;
  StoreActions storeActions; // Multiple "StoreAction"s can be OR-ed
  
  QVector<QPoint> points;
  
private:
  
};

Q_DECLARE_OPERATORS_FOR_FLAGS(Arc::StoreActions)

inline Connectable *Arc::getSource() const {
  return source;
}

inline void Arc::setSource (Connectable *value) {
  source = value;
}

inline Connectable *Arc::getTarget() const {
  return target;
}

inline void Arc::setTarget (Connectable *value) {
  target = value;
}

inline int Arc::getProduction() const {
  return production;
}

inline void Arc::updateProduction(int add) {
  production += add;
}

inline int Arc::getConsumption() const {
  return consumption;
}

inline void Arc::updateConsumption(int add) {
  consumption += add;
}

inline const Arc::Type Arc::getType() const {
  return type;
}

inline void Arc::setType (Arc::Type value) {
  type = value;
}

inline const Arc::StoreActions Arc::getStoreAction() const {
  return storeActions;
}

inline void Arc::setStoreAction (Arc::StoreActions value) {
  storeActions |= value;
}

#endif //ARC_H
