/*
 * FILE:    batchconvertdialog.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "batchconvertdialog.h"
#include "pnmlhandler.h"
#include "export.h"

#include <QFileDialog>
#include <QImageWriter>
#include <QSettings>
#include <QMessageBox>
#include <QFile>

// Pre: dp
BatchConvertDialog::BatchConvertDialog(DrawProps *dp, QWidget * parent /* = 0*/, Qt::WFlags f /* = 0*/) :
  QDialog(parent, f), drawProps(*dp) {
  // Setup UI
  ui.setupUi(this);
  vectorGroupBoxLayout = new QVBoxLayout(ui.vectorGroupBox);
  vectorGroupBoxLayout->setSpacing(6);
  vectorGroupBoxLayout->setMargin(9);
  rasterGroupBoxLayout = new QVBoxLayout(ui.rasterGroupBox);
  rasterGroupBoxLayout->setSpacing(6);
  rasterGroupBoxLayout->setMargin(9);
  ui.tabWidget->hide();
  
  QSettings settings;
  
  // Add export formats
  foreach (QByteArray format, QImageWriter::supportedImageFormats()) {
    if (format.toLower() == "mng" ||
        format.toLower() == "jpeg") {
      // MNG is for multiple pictures and JPG is already included in the
      // list, so JPEG is not needed
      continue;
    } 
    addRasterFormat(QString(format).toUpper(),
        settings.value(QString("BatchConvert/ExportRaster") + QString(format).toUpper(),
                       false).toBool());
  }
#ifndef QT_NO_PRINTER
  addVectorFormat("PDF", settings.value("BatchConvert/ExportVectorPDF", true).toBool());
  addVectorFormat("EPS", settings.value("BatchConvert/ExportVectorEPS", true).toBool());
#endif
#ifdef QT3_SUPPORT
  addVectorFormat("SVG", settings.value("BatchConvert/ExportVectorSVG", false).toBool());
#endif
  vectorGroupBoxLayout->addStretch();
  rasterGroupBoxLayout->addStretch();
  
  // Load settings
  ui.outputDirEdit->setText(settings.value("BatchConvert/OutputDir", QDir::currentPath()).toString());
  // Add a trailing slash to the output dir, if it isn't already there (for the comfort of the LatexPreviewFile default).
  if (!ui.outputDirEdit->text().isEmpty()) {
    if (!ui.outputDirEdit->text().endsWith('/')) {
      ui.outputDirEdit->setText(ui.outputDirEdit->text() + "/");
    }
  }
  lastInputDir = settings.value("BatchConvert/LastInputDir", QDir::currentPath()).toString();
  int zoom = settings.value("BatchConvert/Zoom", 100).toInt();
  if (zoom == 50) {
    ui.zoom50pctRadioButton->setChecked(true);
  } else if (zoom == 100) {
    ui.zoom100pctRadioButton->setChecked(true);
  } else if (zoom == 150) {
    ui.zoom150pctRadioButton->setChecked(true);
  } else if (zoom == 200) {
    ui.zoom200pctRadioButton->setChecked(true);
  } else {
    ui.zoomVarRadioButton->setChecked(true);
    ui.zoomVarSpinBox->setValue(zoom);
  }
  QString overwrite = settings.value("BatchConvert/Overwrite", "always").toString();
  if (overwrite == "ask") {
    ui.overwriteAskRadioButton->setChecked(true);
  } else if (overwrite == "never") {
    ui.overwriteNeverRadioButton->setChecked(true);
  } else {
    ui.overwriteAlwaysRadioButton->setChecked(true);
  }
  ui.exportPagesCheckBox->setChecked(settings.value("BatchConvert/ExportPages", true).toBool());
  ui.subnetDelimEdit->setText(settings.value("BatchConvert/SubnetDelimiter", "_").toString());
  ui.subnetDelimEdit->setEnabled(ui.exportPagesCheckBox->isChecked());
  ui.subnetDelimLabel->setEnabled(ui.exportPagesCheckBox->isChecked());
  
  // LaTeX settings
  ui.latexCheckBox->setChecked(settings.value("BatchConvert/LatexPreview", false).toBool());
  
  ui.latexFileEdit->setText(settings.value("BatchConvert/LatexPreviewFile", ui.outputDirEdit->text() + "preview.tex").toString());
  ui.textWidthCheckBox->setChecked(settings.value("BatchConvert/LatexTextWidth", false).toBool());
  ui.printDescriptionsCheckBox->setChecked(settings.value("BatchConvert/PrintDescriptions", false).toBool());
  ui.printQueriesCheckBox->setChecked(settings.value("BatchConvert/PrintQueries", false).toBool());
  ui.printRolesCheckBox->setChecked(settings.value("BatchConvert/PrintRoles", false).toBool());
  ui.printPrePostsetsCheckBox->setChecked(settings.value("BatchConvert/PrintPrePostsets", false).toBool());
  
  ui.latexFileEdit->setEnabled(ui.latexCheckBox->isChecked());
  ui.browseLatexButton->setEnabled(ui.latexCheckBox->isChecked());
  ui.transitionGroupBox->setEnabled(ui.latexCheckBox->isChecked());
}

BatchConvertDialog::~BatchConvertDialog() {
}

void BatchConvertDialog::addVectorFormat(QString format, bool checked /*= false*/) {
  QCheckBox *cb = new QCheckBox(ui.vectorGroupBox);
  cb->setText(format);
  cb->setChecked(checked);
  vectorGroupBoxLayout->addWidget(cb);
  vectorCheckBoxes << cb;
}

void BatchConvertDialog::addRasterFormat(QString format, bool checked /*= false*/) {
  QCheckBox *cb = new QCheckBox(ui.rasterGroupBox);
  cb->setText(format);
  cb->setChecked(checked);
  rasterGroupBoxLayout->addWidget(cb);
  rasterCheckBoxes << cb;
}

void BatchConvertDialog::on_addFilesButton_clicked() {
  QStringList files = QFileDialog::getOpenFileNames(
                            this,
                            "Select one or more files to open",
                            lastInputDir,
                            "PNML files (*.pnml)");
  QString file;
  foreach (file, files) {
    new QListWidgetItem(file, ui.fileList);
  }
  if (!file.isEmpty()) {
    QFileInfo fi(file);
    lastInputDir = fi.absolutePath();
  }
}

void BatchConvertDialog::on_browseButton_clicked() {
  QString dir = QFileDialog::getExistingDirectory(
                            this,
                            "Select an output directory",
                            ui.outputDirEdit->text());
  if (!dir.isEmpty()) {
    ui.outputDirEdit->setText(dir);
  }
}

void BatchConvertDialog::on_browseLatexButton_clicked() {
  QString fileName = QFileDialog::getSaveFileName(this,
                                                  "Save LaTeX preview",
                                                  ui.latexFileEdit->text(),
                                                  "LaTeX file (*.tex)");
  if (!fileName.isEmpty()) {
    ui.latexFileEdit->setText(fileName);
  }
}

void BatchConvertDialog::on_convertButton_clicked() {
  // Do conversion
  ui.progressBar->setValue(0);
  ui.outputLog->clear();
  
  // Check if output dir exists, if not, ask if we should create it.
  QDir outputDir(ui.outputDirEdit->text());
  if (!outputDir.exists()) {
    int reply = QMessageBox::question(this,
                                      "Output directory",
                                      "Output directory doesn't exist, shall I create it?",
                                      QMessageBox::Yes,
                                      QMessageBox::No);
    if (reply == QMessageBox::Yes) {
      if (!outputDir.mkpath(ui.outputDirEdit->text())) {
        QMessageBox::warning(this,
                             "Cannot create directory",
                             "The output directory could not be created.");
        return;
      }
    } else {
      return;
    }
  }
  // Add a trailing slash to the output dir, if it isn't already there.
  Q_ASSERT(!ui.outputDirEdit->text().isEmpty()); // shouldn't be empty because it is a valid directory
  if (!ui.outputDirEdit->text().endsWith('/')) {
    ui.outputDirEdit->setText(ui.outputDirEdit->text() + "/");
  }
  
  // Open LaTeX output file
  QFile texFile;
  if (ui.latexCheckBox->isChecked()) {
    QFileInfo fi(ui.latexFileEdit->text());
    if (fi.exists()) {
      if (ui.overwriteAskRadioButton->isChecked()) {
        if (QMessageBox::question(this,
                                    "LaTeX file exists",
                                    QString("LaTeX output file ``%1'' already exists, Do you want to replace it?").arg(ui.latexFileEdit->text()),
                                    QMessageBox::Yes,
                                    QMessageBox::No)
            == QMessageBox::No) {
          ui.outputLog->append(QString("<font color=\"red\">Not saving LaTeX preview to ``%1'', file already exists. Please disable LaTeX preview, change the output file or accept overwrite.</font>")
                                      .arg(ui.latexFileEdit->text()));
          return;
        }
      } else if (ui.overwriteNeverRadioButton->isChecked()) {
        ui.outputLog->append(QString("<font color=\"red\">Not saving LaTeX preview to ``%1'', file already exists. Please disable LaTeX preview, change the output file or accept overwrite.</font>")
                                    .arg(ui.latexFileEdit->text()));
        return;
      }
    }
    texFile.setFileName(ui.latexFileEdit->text());
    if (texFile.open(QFile::WriteOnly | QFile::Truncate)) {
      latexStream.setDevice(&texFile);
      
      // Write LaTeX header
      latexStream << "\\documentclass[10pt]{article}" << endl;
      latexStream << "\\usepackage{a4wide}" << endl;
      latexStream << "\\usepackage{graphicx}" << endl;
      if (ui.printQueriesCheckBox->isChecked()) {
        latexStream << "\\usepackage{listings}" << endl << endl;
        latexStream << "\\lstset{language=SQL,morekeywords={TRUNCATE},sensitive=true}" << endl;
        // TODO: add some functions to be "highlighted":
        //   IF(), NOW(), CONCAT()
      }
      latexStream << endl << "\\begin{document}" << endl;
    } else {
      ui.outputLog->append(QString("<font color=\"red\">Couldn't open LaTeX output file: ``%1''</font>")
                            .arg(ui.latexFileEdit->text()));
      return;
    }
  }
  
  // Apply zoom
  if (ui.zoom50pctRadioButton->isChecked()) {
    drawProps.setZoom(0.5);
  } else if (ui.zoom100pctRadioButton->isChecked()) {
    drawProps.setZoom(1.0);
  } else if (ui.zoom150pctRadioButton->isChecked()) {
    drawProps.setZoom(1.5);
  } else if (ui.zoom200pctRadioButton->isChecked()) {
    drawProps.setZoom(2.0);
  } else {
    drawProps.setZoom(1.0 * ui.zoomVarSpinBox->value() / 100.0);
  }
  
  // Convert all files
  for (int i = 0; i < ui.fileList->count(); i++) {
    ui.outputLog->append(QString("<b>") + ui.fileList->item(i)->text() + "</b>");
    
    // Create a handler for the file
    PnmlHandler handler;
    QXmlSimpleReader reader;
    reader.setContentHandler(&handler);
    reader.setErrorHandler(&handler);

    // Test if the file can be opened
    QFile file(ui.fileList->item(i)->text());
    if (!file.open(QFile::ReadOnly | QFile::Text)) {
      // TODO write this message to ui.outputLog instead of messagebox
      QMessageBox::warning(this, "PNML view",
                           QString("Cannot read file %1:\n%2.\n\nI will continue processing other files.")
                           .arg(ui.fileList->item(i)->text())
                           .arg(file.errorString()));
      continue;
    }

    // Start the parsing procedure
    QXmlInputSource xmlInputSource(&file);
    if (reader.parse(xmlInputSource)) {
      // File parsed and ok
      ui.outputLog->append("Parse ok");
      
      // Export to all checked formats
      QFileInfo fi(ui.fileList->item(i)->text());
      if (handler.getNet()) {
        recurseExportNet(handler.getNet(), fi.baseName());
      } else {
        ui.outputLog->append("No net found in input file");
      }
    } else {
      ui.outputLog->append(
          QString("<font color=\"red\">Parsing of input file failed: %1.</font>")
            .arg(handler.fatalErrorString()));
    }
    
    // Update progress bar
    ui.progressBar->setValue((i+1)*100/ui.fileList->count());
    QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
  }
  
  // Close LaTeX output file
  if (ui.latexCheckBox->isChecked()) {
    latexStream << "\\end{document}" << endl;
    
    texFile.close();
  }
  
  ui.progressBar->setValue(100);
  ui.outputLog->append("<b>Done!</b>");
}

// Recursively export a net, starting with the net itself and going through
// its hierarchy
// Pre: net
void BatchConvertDialog::recurseExportNet(Net *net, QString baseFileName) {
  // Don't export empty nets (such as inline pages)
  if (net->isEmpty()) {
    return;
  }
  
  // Export current net
  exportNet(net, baseFileName);
  
  // Add net to LaTeX output file
  if (ui.latexCheckBox->isChecked()) {
    bool dumpAny = ui.printDescriptionsCheckBox->isChecked() ||
                   ui.printQueriesCheckBox->isChecked() ||
                   ui.printRolesCheckBox->isChecked() ||
                   ui.printPrePostsetsCheckBox->isChecked();
    
    // Determine relative filepath from LaTeX preview file to graphic
    QFileInfo fi(ui.latexFileEdit->text());
    QString baseName = 
        fi.absoluteDir()
          .relativeFilePath(ui.outputDirEdit->text() +
                            replaceSpecialPathChars(baseFileName));
    
    QString caption = latexFormat(baseFileName, true);
    if (dumpAny) {
      latexStream << "\\section{" << caption << "}\n";
    }
    latexStream << "\\begin{figure}[h]" << endl;
    latexStream << "  \\begin{center}" << endl;
    latexStream << "    \\includegraphics";
    if (ui.textWidthCheckBox->isChecked()) {
      latexStream << "[width=\\textwidth]";
    }
    latexStream << "{" << baseName << "}" << endl;
    latexStream << "    \\caption{" << caption << "}" << endl;
    latexStream << "  \\end{center}" << endl;
    latexStream << "\\end{figure}" << endl;
    if (dumpAny) {
      Connectable *curCon;
      foreach (curCon, net->getConnectables()) {
        if (curCon->getConType() == Connectable::Transition) {
          Transition *curTrans = dynamic_cast<Transition *>(curCon);
          if (curTrans) {
            QString description;

            // Print description
            if (ui.printDescriptionsCheckBox->isChecked() &&
                !curCon->getDescription().isEmpty()) {
              description += latexFormat(curCon->getDescription()) + "\\\\\n";
            }

            // Print pre- & postsets
            if (ui.printQueriesCheckBox->isChecked()) {
              if (!curTrans->getPreQuery().isEmpty()) {
                description += "{\\bf Prequery}\n";
                description += "\\begin{lstlisting}[breaklines=true,breakatwhitespace=true,columns=flexible]\n";
                description += curTrans->getPreQuery() + "\n";
                description += "\\end{lstlisting}\n";
              }
              if (!curTrans->getPostQuery().isEmpty()) {
                description += "{\\bf Postquery}\n";
                description += "\\begin{lstlisting}[breaklines=true,breakatwhitespace=true,columns=flexible]\n";
                description += curTrans->getPostQuery() + "\n";
                description += "\\end{lstlisting}\n";
              }
            }
            
            // Print pre- & postquery
            if (ui.printPrePostsetsCheckBox->isChecked()) {
              description += setJoin(curCon->getPreSet(), "Pre-set", ", ");
              description += setJoin(curCon->getPostSet(), "Post-set", ", ");
              description += setJoin(curCon->getAssocStores(), "Stores", ", ");
              description += setJoin(curCon->getInhibits(), "Inhibited by", ", ");
              description += setJoin(curCon->getResets(), "Resets", ", ");
            }

            // Print role(s)
            if (ui.printRolesCheckBox->isChecked() &&
                !curTrans->getRoles().isEmpty()) {
              description += "{\\bf Roles} ";
              description += latexFormat(curTrans->getRoles().join(", "));
              description += "\\\\\n";
            }
            
            if (!description.isEmpty()) {
              latexStream << "\\subsection{"
                          << latexFormat(curCon->getNameOrId(), true)
                          << "}\n"
                          << description;
            }
          }
        }
      } 
    }
    latexStream << "\\clearpage" << endl << endl;
  }
  
  if (ui.exportPagesCheckBox->isChecked()) {
    // Export child nets
    Page *curPage;
    foreach (curPage, net->getPages()) {
      if (curPage->getNet()) {
        QString pageName = curPage->getNameOrId();
        ui.outputLog->append("Exporting page ``" + pageName + "'':");
        recurseExportNet(curPage->getNet(),
                         baseFileName +
                         ui.subnetDelimEdit->text() +
                         pageName);
      }
    }
  }
}


// Pre: net
void BatchConvertDialog::exportNet(Net *net, QString baseName) {
  baseName = ui.outputDirEdit->text() + replaceSpecialPathChars(baseName);
  QCheckBox *cb;
  foreach (cb, vectorCheckBoxes) {
    if (cb->isChecked()) {
      bool exportOk = true;
      
      QString fileName = baseName + "." + cb->text().toLower();
      QFileInfo fi(fileName);
      if (fi.exists()) {
        if (ui.overwriteAskRadioButton->isChecked()) {
          if (QMessageBox::question(this,
                                      "File exists",
                                      QString("Output file ``%1'' already exists, Do you want to replace it?").arg(fileName),
                                      QMessageBox::Yes,
                                      QMessageBox::No)
              == QMessageBox::No) {
            ui.outputLog->append(QString("Not saving to ``%1'', file already exists.")
                                        .arg(fileName));
            continue;
          }
        } else if (ui.overwriteNeverRadioButton->isChecked()) {
          ui.outputLog->append(QString("Not saving to ``%1'', file already exists.")
                                      .arg(fileName));
          continue;
        }
      }
      if (cb->text() == "PDF") {
        exportOk &= Export::savePdf(net, &drawProps, fileName);
      } else if (cb->text() == "EPS") {
        exportOk &= Export::saveEps(net, &drawProps, fileName);
      } else if (cb->text() == "SVG") {
        exportOk &= Export::saveSvg(net, &drawProps, fileName);
      }
      
      if (exportOk) {
        ui.outputLog->append(QString("Saved as ") + cb->text());
      } else {
        ui.outputLog->append(QString("<font color=\"red\">File ``%1'' could not be written.</font>")
                                    .arg(fileName));
      }
      QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
    }
  }
  foreach (cb, rasterCheckBoxes) {
    if (cb->isChecked()) {
      bool exportOk = true;
      QString fileName = baseName + "." + cb->text().toLower();
      QFileInfo fi(fileName);
      if (fi.exists()) {
        if (ui.overwriteAskRadioButton->isChecked()) {
          if (QMessageBox::question(this,
                                      "File exists",
                                      QString("Output file ``%1'' already exists, Do you want to replace it?").arg(fileName),
                                      QMessageBox::Yes,
                                      QMessageBox::No)
              == QMessageBox::No) {
            ui.outputLog->append(QString("Not saving to ``%1'', file already exists.")
                                        .arg(fileName));
            continue;
          }
        } else if (ui.overwriteNeverRadioButton->isChecked()) {
          ui.outputLog->append(QString("Not saving to ``%1'', file already exists.")
                                      .arg(fileName));
          continue;
        }
      }
      exportOk &= Export::saveRaster(net,
                         &drawProps,
                         fileName,
                         cb->text());
      if (exportOk) {
        ui.outputLog->append(QString("Saved as ") + cb->text());
      } else {
        ui.outputLog->append(QString("<font color=\"red\">File ``%1'' could not be written.</font>")
                                    .arg(fileName));
      }
      QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
    }
  }
}

void BatchConvertDialog::on_closeButton_clicked() {
  // Save settings
  QSettings settings;
  
  QCheckBox *cb;
  foreach (cb, vectorCheckBoxes) {
    settings.setValue(
        QString("BatchConvert/ExportVector") + cb->text(),
        cb->isChecked());
  }
  foreach (cb, rasterCheckBoxes) {
    settings.setValue(
        QString("BatchConvert/ExportRaster") + cb->text(),
        cb->isChecked());
  }
  
  settings.setValue("BatchConvert/OutputDir", ui.outputDirEdit->text());
  settings.setValue("BatchConvert/LastInputDir", lastInputDir);
  if (ui.zoom50pctRadioButton->isChecked()) {
    settings.setValue("BatchConvert/Zoom", 50);
  } else if (ui.zoom100pctRadioButton->isChecked()) {
    settings.setValue("BatchConvert/Zoom", 100);
  } else if (ui.zoom150pctRadioButton->isChecked()) {
    settings.setValue("BatchConvert/Zoom", 150);
  } else if (ui.zoom200pctRadioButton->isChecked()) {
    settings.setValue("BatchConvert/Zoom", 200);
  } else {
    settings.setValue("BatchConvert/Zoom", ui.zoomVarSpinBox->value());
  }
  if (ui.overwriteAlwaysRadioButton->isChecked()) {
    settings.setValue("BatchConvert/Overwrite", "always");
  } else if (ui.overwriteAskRadioButton->isChecked()) {
    settings.setValue("BatchConvert/Overwrite", "ask");
  } else if (ui.overwriteNeverRadioButton->isChecked()) {
    settings.setValue("BatchConvert/Overwrite", "never");
  }
  settings.setValue("BatchConvert/ExportPages", ui.exportPagesCheckBox->isChecked());
  settings.setValue("BatchConvert/SubnetDelimiter", ui.subnetDelimEdit->text());
  
  // LaTeX settings
  settings.setValue("BatchConvert/LatexPreview", ui.latexCheckBox->isChecked());
  settings.setValue("BatchConvert/LatexPreviewFile", ui.latexFileEdit->text());
  settings.setValue("BatchConvert/LatexTextWidth", ui.textWidthCheckBox->isChecked());
  settings.setValue("BatchConvert/PrintDescriptions", ui.printDescriptionsCheckBox->isChecked());
  settings.setValue("BatchConvert/PrintQueries", ui.printQueriesCheckBox->isChecked());
  settings.setValue("BatchConvert/PrintRoles", ui.printRolesCheckBox->isChecked());
  settings.setValue("BatchConvert/PrintPrePostsets", ui.printPrePostsetsCheckBox->isChecked());
  
  // Hide the dialog
  accept();
}

QString BatchConvertDialog::replaceSpecialPathChars(QString str) {
  str.replace(' ', '_');
  str.replace('/', '_');
  str.replace('\\', '_');
  str.replace(':', '_');
  str.replace('\n', '_');
  return str;
}

QString BatchConvertDialog::latexFormat(QString str, bool noEndl /*=false*/) {
  str = str.trimmed();
  str.replace('\\', "\\textbackslash"); // should go first!
  str.replace('_', "\\_");
  str.replace('{', "\\{");
  str.replace('}', "\\}");
  str.replace('$', "\\$");
  str.replace('#', "\\#");
  str.replace('&', "\\&");
  str.replace('<', "\\textless ");
  str.replace('>', "\\textgreater ");
  str.replace('|', "\\textbar ");
  str.replace('~', "\\textasciitilde ");
  str.replace('^', "\\textasciicircum ");
  if (noEndl) {
    str.replace('\n', ' ');
  } else {
    str.replace('\n', "\\\\\n");
  }
  return str;
}

const QString BatchConvertDialog::setJoin(const QVector<Connectable *> set, const QString setName, const QString join) {
  QString str;
  if (!set.isEmpty()) {
    str += "{\\bf " + setName + "} ";
    
    bool first = true;
    foreach (Connectable *con, set) {
      str += (!first?join:"") + latexFormat(con->getNameOrId());
      first = false;
    }
    
    str += "\\\\\n";
  }
  return str;
}

