/*
 * FILE:    drawwidget.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "drawwidget.h"
#include "drawprops.h"
#include "page.h"
#include "place.h"
#include "net.h"
#include "arc.h"

#include <QPainter>
#include <QMouseEvent>
#include <QToolTip>

#define UNIT 50
#define WHEEL_DELTA_SCALE (8.0 * 120)
#define ZOOM_SCALE 50.0

// Pre: dp
DrawWidget::DrawWidget(DrawProps *dp, QWidget * parent /* = 0*/, Qt::WFlags f /* = 0*/) :
  QWidget(parent, f), drawProps(dp) {
  net = NULL;
  
  showToolTips = false;
  toolTipTimer.setSingleShot(true);
  connect(&toolTipTimer, SIGNAL(timeout()), this, SLOT(showToolTip()));
  setAutoFillBackground(false);
  setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
  resize(0, 0);
  setMouseTracking(true);
}

DrawWidget::~DrawWidget() {
}

void DrawWidget::setNet(Net *value) {
  if (net != value) {
    net = value;
    sizeChanged();
  }
}

Net *DrawWidget::getNet() const {
  return net;
}

void DrawWidget::paintEvent(QPaintEvent *e) {
  if (net) {
    drawProps->updateRegion = e->region();
    net->draw(this, drawProps);
    drawProps->updateRegion = QRegion();
  }
}

QSize DrawWidget::sizeHint() const {
  return size();
}


void DrawWidget::sizeChanged() {
  setCursor(Qt::ArrowCursor);
  if (net) {
    resize(net->getSize(drawProps));
  }
  update();
  
  // TODO: Should we call updateGeometry() here? It appears to work without...
}

void DrawWidget::mouseReleaseEvent(QMouseEvent *e) {
  if (net) {
    bool found = false;
    QPoint pos = net->demapPos(e->pos(), drawProps);
    
    Page *curPage;
    foreach (curPage, net->getPages()) {
      if (curPage->getType() != Transition::InlinePage &&
          curPage->containsPos(pos)) {
        net = curPage->getNet();
        sizeChanged();
        setCursor(Qt::ArrowCursor);
        QToolTip::showText(QPoint(), "", this);
        // TODO: emit signal to let mainwindow update hierarchy selection
        found = true;
        break;
      }
    }
    
    if (!found) {
      Connectable *curCon;
      foreach (curCon, net->getConnectables()) {
        if (curCon->getConType() == Connectable::Place) {
          Place *curPlace = dynamic_cast<Place *>(curCon);
          if (curPlace &&
              curPlace->getType() == Place::Reference &&
              curPlace->containsPos(pos)) {
            net = net->getParent();
            sizeChanged();
            setCursor(Qt::ArrowCursor);
            QToolTip::showText(QPoint(), "", this);
            // TODO: emit signal to let mainwindow update hierarchy selection
            found = true;
            break;
          }
        }
      }
    }
  }
  e->accept();
}

void DrawWidget::mouseMoveEvent(QMouseEvent *e) {
  bool found = false;
  if (net) {
    QPoint pos = net->demapPos(e->pos(), drawProps);
    
    Page *curPage;
    foreach (curPage, net->getPages()) {
      if (curPage->getType() != Transition::InlinePage &&
          curPage->containsPos(pos)) {
        setCursor(Qt::PointingHandCursor);
        found = true;
        break;
      }
    }
    if (!found) {
      foreach (Connectable *curCon, net->getConnectables()) {
        if (curCon->getConType() == Connectable::Place) {
          Place *curPlace = dynamic_cast<Place *>(curCon);
          if (curPlace &&
              curPlace->getType() == Place::Reference &&
              curPlace->containsPos(pos)) {
            setCursor(Qt::PointingHandCursor);
            found = true;
            break;
          }
        }
      }
    }
    if (!found) {
      setCursor(Qt::ArrowCursor);
    }
    
    if (showToolTips) {
      lastMousePos = e->pos();
      toolTipTimer.start(100);
    }
  }
  e->accept();
}

void DrawWidget::showToolTip() {
  QPoint pos = net->demapPos(lastMousePos, drawProps);
  bool found = false;
  if (!found) {
    foreach (Connectable *curCon, net->getConnectables()) {
      if (curCon->containsPos(pos)) {
        QToolTip::showText(mapToGlobal(lastMousePos), curCon->getToolTip(), this);
        found = true;
        break;
      }
    }
  }
  if (!found) {
    foreach (Arc *curArc, net->getArcs()) {
      if (curArc->containsPos(pos)) {
        QToolTip::showText(mapToGlobal(lastMousePos), curArc->getToolTip(), this);
        found = true;
        break;
      }
    }
  }
  if (!found) {
    QToolTip::showText(QPoint(), "", this);
  }

}
