/*
 * FILE:    export.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "export.h"
#include "drawprops.h"
#include "net.h"

#include <QFile>
#include <QFileInfo>
#include <QTextStream>
#include <QRegExp>
#include <QImage>
#ifdef QT3_SUPPORT
#include <Q3Picture>
#endif

//--- begin my pdf print engine
#ifdef MY_PDF
#include "pdf/qprintengine_pdf_p.h"
#endif
#include <QPrinter>
//--- end my pdf print engine

#include <math.h>

// Pre: net && drawProps
bool Export::savePdf(Net *net, DrawProps *drawProps, const QString fileName) {
  bool exportOk = true;
#ifndef QT_NO_PRINTER
  QPrinter printer(QPrinter::ScreenResolution);
  printer.setOutputFileName(fileName);
#if QT_VERSION >= 0x040100
  printer.setOutputFormat(QPrinter::PdfFormat);
#endif

#ifdef MY_PDF
  printer.setFullPage(true);
  QPdfEngine *pdfEngine = dynamic_cast<QPdfEngine *>(printer.printEngine());
  if (pdfEngine) {
    pdfEngine->setPaperSize(net->getSize(drawProps));
  }
#endif

  DrawProps dp(*drawProps);
  dp.setPrintSafe();
#ifndef MY_PDF
  // Scale to paper
  QSize netSize = net->getSize(&dp);
  if (1.0 * netSize.width() / netSize.height() >
      1.0 * printer.width() / printer.height()) {
    // Net is wider than paper
    dp.setZoom(1.0 * printer.width() / netSize.width());
  } else {
    // Net is higher than paper
    dp.setZoom(1.0 * printer.height() / netSize.height());
  }
#endif
  net->draw(&printer, &dp, true);

  exportOk = QFile::exists(fileName);
#else
  exportOk = false;
#endif
  return exportOk;
}

// Pre: net && drawProps
bool Export::saveEps(Net *net, DrawProps *drawProps, const QString fileName) {
  bool exportOk = true;
#ifndef QT_NO_PRINTER
  
  QPrinter *printer = new QPrinter(QPrinter::ScreenResolution);
  Q_ASSERT(printer);
  printer->setOutputFileName(fileName);
#if QT_VERSION >= 0x040200
  printer->setOutputFormat(QPrinter::PostScriptFormat);
#endif
  printer->setColorMode(QPrinter::Color);
  printer->setFullPage(true);

  DrawProps dp(*drawProps);
  dp.setPrintSafe();
  net->draw(printer, &dp, true);

  int resolution = printer->resolution();
  if (resolution < 1) {
    resolution = 72; // some default, just in case resolution call fails
  }
  
  // delete printer before we try to open and fix the file
  delete printer;

  // modify bounding box from screen to eps resolution.
  QSize size = net->getSize(drawProps);
  size.setWidth((int)ceil(size.width() * 72.0/resolution));
  size.setHeight((int)ceil(size.height() * 72.0/resolution));
  exportOk &= fixEps(fileName, size);
#else
  exportOk = false;
#endif
  return exportOk;
}

// Pre: net && drawProps
bool Export::saveSvg(Net *net, DrawProps *drawProps, const QString fileName) {
#ifdef QT3_SUPPORT
  Q3Picture pict;

  DrawProps dp(*drawProps);
  dp.setPrintSafe();
  net->draw(&pict, &dp, true);

  return pict.save(fileName, "svg");
#else
  return false;
#endif
}

// Pre: net && drawProps
bool Export::saveRaster(Net *net, DrawProps *drawProps, const QString fileName, QString format) {
  bool exportOk = true;
  
  format = format.toLower();
  QImage *image;
  if (format == "png" || format == "gif" || format == "xpm") {
    // Formats that are able to handle transparency
    image = new QImage(net->getSize(drawProps), QImage::Format_ARGB32_Premultiplied);
  } else {
    // Format is not able to handle transparancy, so make background white
    image = new QImage(net->getSize(drawProps), QImage::Format_RGB32);
  }
  Q_ASSERT(image);
  if (format == "png" && drawProps->option("ExportTransparency")) {
    // PNG supports alpha-transparency :-D
    DrawProps dp(*drawProps);
    dp.setBrush("Background", Qt::NoBrush);
    net->draw(image, &dp, true);
  } else {
    net->draw(image, drawProps);
  }

  exportOk &= image->save(fileName, qPrintable(format));
  delete image;
  
  return exportOk;
}

// Function from umbrello (uml.sourceforge.net), UMLView::fixEPS()
//  (but translated to Qt4)
bool Export::fixEps(const QString &filename, QSize size) {
  // now open the file and make a correct eps out of it
  QFile epsfile(filename);
  if (!epsfile.open(QIODevice::ReadOnly)) {
    return false;
  }
  // read
  QTextStream ts(&epsfile);
  QString fileContent = ts.readAll();
  epsfile.close();

  // read information
  QRegExp rx("%%BoundingBox:\\s*(-?[\\d\\.:]+)\\s*(-?[\\d\\.:]+)\\s*(-?[\\d\\.:]+)\\s*(-?[\\d\\.:]+)");
  const int pos = rx.indexIn(fileContent);
  if (pos < 0) {
    qCritical("Export::fixEPS(%s): cannot find %%BoundingBox", qPrintable(filename));
    return false;
  }

  // write new content to file
  if (!epsfile.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
    qCritical("Export::fixEPS(%s): cannot open file for writing", qPrintable(filename));
    return false;
  }

  // be careful when rounding (ceil/floor) the BB, these roundings
  // were mainly obtained experimentally...
  const double epsleft = rx.cap(1).toFloat();
  const double epstop = rx.cap(4).toFloat();
  const int left = int(floor(epsleft));
  const int right = int(ceil(epsleft)) + size.width();
  const int top = int(ceil(epstop)) + 1;
  const int bottom = int(floor(epstop)) - size.height() + 1;

  // modify content
  fileContent.replace(pos,rx.cap(0).length(),
                      QString("%%BoundingBox: %1 %2 %3 %4").arg(left).arg(bottom).arg(right).arg(top));

  ts << fileContent;
  epsfile.close();
  
  return true;
}

