/*
 * FILE:    place.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "place.h"
#include "labelplacer.h"
#include "drawprops.h"

#include <QPainter>
#include <math.h>
#include "defs.h"

Place::Place() : Connectable() {
  conType = Connectable::Place;
  type = Channel;
  initialMarking = 0;
  caseSensitive = TokenCaseInsensitive;
  refType = None;
  reference = NULL;
#ifdef SHOW_FAKE_ARCS
  pathGraphics = NULL;
#endif
  
  // Default size (if none is given in input file)
  size = QSize(20, 20);
}

Place::~Place() {
#ifdef SHOW_FAKE_ARCS
  delete pathGraphics;
#endif
}

// Pre: p.isActive() && drawProps
void Place::draw (QPainter *p, DrawProps *drawProps) {
//  qDebug("Place::draw() id=%s", qPrintable(id));
  
  // Make rectangle
  QRectF rect(1.0*position.x() - .5*size.width(), 
              1.0*position.y() - .5*size.height(),
              size.width(),
              size.height());
  
  // Pen used to draw everything
  QPen pen(drawProps->pen("Default"));
  
  if (type == Store) {
    // Draw a store symbol (database)
    
    // Ellipse rectangle (for top & bottom)
    QRectF ellipse(rect);
    ellipse.setHeight(1.0*size.height()/3);
    // Body rectangle
    QRectF bodyRect(rect);
    bodyRect.setHeight(1.0*size.height()*2/3);
    bodyRect.translate(0, 1.0*size.height()/6);
    
    // Draw shadow
    p->setPen(Qt::NoPen); 
    p->setBrush(drawProps->brush("Shadow"));
    
    // Draw bottom ellipse (shadow)
    p->drawChord(ellipse.translated(0, 1.0*size.height()*2/3).translated(SHADOW_SIZE,SHADOW_SIZE), 2880, 2880);
    
    // Draw body (shadow)
    p->drawRect(bodyRect.translated(SHADOW_SIZE,SHADOW_SIZE));
    
    // Draw upper ellipse (shadow)
    p->drawChord(ellipse.translated(SHADOW_SIZE,SHADOW_SIZE), 0, 2880);
    
    
    // Draw symbol
    p->setPen(pen);
    if (caseSensitive == TokenCaseSensitive) {
      p->setBrush(drawProps->brush("ColoredStore"));
    } else {
      p->setBrush(drawProps->brush("UncoloredStore"));
    }
    
    // Draw bottom ellipse (symbol)
    p->drawEllipse(ellipse.translated(0, 1.0*size.height()*2/3));
    
    // Draw body (symbol)
    p->setPen(Qt::NoPen);
    p->drawRect(bodyRect);
    p->setPen(pen);
    p->drawLine(bodyRect.bottomLeft(), bodyRect.topLeft());
    p->drawLine(bodyRect.bottomRight(), bodyRect.topRight());
    
    // Draw upper ellipse (symbol)
    p->drawEllipse(ellipse);
  } else { // Normal place
    // Draw shadow
    p->setPen(Qt::NoPen);
    p->setBrush(drawProps->brush("Shadow"));
    p->drawEllipse(rect.translated(SHADOW_SIZE,SHADOW_SIZE));

    // Draw place
    p->setPen(pen);
    if (caseSensitive == TokenCaseSensitive) {
      p->setBrush(drawProps->brush("ColoredPlace"));
    } else {
      p->setBrush(drawProps->brush("UncoloredPlace"));
    }
    p->drawEllipse(rect);
    
    // Draw characters for reference arcs:
    //  "i" = inflow
    //  "o" = outflow
    //  "x" = inhibitor
    //  "r" = reset
    //  "b" = biflow
    // btw: due to the limitations of pnml, only one of these types
    //      is active
    if (type == Reference && refType != None) {
      QFont font = drawProps->font("Name");
      font.setPixelSize(size.height()/2);
      font.setWeight(QFont::Bold);
      p->setFont(font);
      QTextOption opt;
      opt.setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
      QString letter;
      switch (refType) {
        case InFlowRef:
          letter = "i";
          break;
        case OutFlowRef:
          letter = "o";
          break;
        case BiFlowRef:
          letter = "b";
          break;
        case InhibitorRef:
          letter = "x";
          break;
        case ResetRef:
          letter = "r";
          break;
        case StoreRef:
          letter = "s";
          break;
        case None:
          break;
      }
      p->drawText(rect, letter, opt);
    }
  }
  
  // Draw tokens
  if (initialMarking > 0) {
    QPointF center(position);
    double circleRadius = size.width() / 3;
    double tokenRadius = 1.5;
    QVector<QPointF> tokens;
    int tokensToDraw = initialMarking;
    
    if (tokensToDraw < 9) {
      // Draw center token
      if (tokensToDraw == 1 || tokensToDraw > 4) {
        tokens << QPointF(0, 0);
        tokensToDraw--;
      }

      // Draw tokens in circle
      if (tokensToDraw > 0) {
        double radPerToken = 2 * M_PI / tokensToDraw;
        for (int i = 0; i < tokensToDraw; i++) {
          tokens << QPointF(cos(i * radPerToken),
                            sin(i * radPerToken));
        }
      }
    } else if (initialMarking < 100) {
      // initialMarking >= 9
      p->setPen(drawProps->pen("Text"));
      p->setFont(drawProps->font("Name"));

      QTextOption opt;
      opt.setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
      opt.setWrapMode(QTextOption::NoWrap);
  
      p->drawText(rect, QString::number(initialMarking), opt);
    } /* else {
      // Number of tokens is added as label in updateLabels() 
    } */
    
    p->setPen(pen);
    p->setBrush(drawProps->brush("Token"));
    foreach (QPointF token, tokens) {
      token *= circleRadius;
      token += position - QPointF(tokenRadius, tokenRadius);
      p->drawEllipse(QRectF(token, QSizeF(tokenRadius * 2, tokenRadius * 2)));
    }
  }
}

// Pre: labelPlacer && drawProps
void Place::updateLabels(LabelPlacer *labelPlacer, DrawProps *drawProps) {
  labelPlacer->addRect(QRect(position.x() - size.width()/2,
                             position.y() - size.height()/2,
                             size.width(),
                             size.height()));
  if (!name.isEmpty()) {
    labelPlacer->addLabel(position, name);
  } else if (type == Reference && reference &&
             !reference->getName().isEmpty()) {
    labelPlacer->addLabel(position, reference->getName());
  } else if (drawProps->option("ShowIdsPlaces")) {
    labelPlacer->addLabel(position, id);
  }
  
  if (initialMarking >= 100) {
    labelPlacer->addLabel(position, QString::number(initialMarking));
  }
}

#define SIGN(x) ((x)<0?-1:1)

const QPointF Place::getConnectionPoint(QPointF source) const {
  if (type == Store) {
    QPointF p(source - position);
    if (fabs(p.x()) > fabs(p.y())) {
      p.setX(SIGN(p.x()) * size.width()/2);
      p.setY(0);
    } else {
      p.setX(0);
      p.setY(SIGN(p.y()) * size.height()/2);
    }
    return p + position;
  } else {
    QLineF line(position, source);
    // TODO: change next line to be able to handle ellipses (size.height() != size.width())
    line.setLength(0.5 * size.height());
    return line.p2();
  }
}

const QRect Place::getBoundingRect() const {
  QRect rect(position.x() - size.width()/2,
             position.y() - size.height()/2,
             size.width() + SHADOW_SIZE,
             size.height() + SHADOW_SIZE);
  return rect;
}

const QString Place::getToolTip() const {
  QString text = Connectable::getToolTip();
  if (type == Reference && reference) {
    if (!reference->getName().isEmpty()) {
      text += markup(reference->getName() + " (" + reference->getId() + ")", "Reference to:");
    } else {
      text += markup(reference->getId(), "Reference to:");
    }
  }
  if (initialMarking > 0) {
    if (initialMarking == 1) {
      text += markup("1 token", "Initial marking:");
    } else {
      text += markup(QString::number(initialMarking) + " tokens", "Initial marking:");
    }
  }
  return text;
}
