/*
 * FILE:    prefdialog.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "prefdialog.h"
#include "drawprops.h"

#include <QDateTime>
#include <QFontDialog>

#include <cstdlib>

class QRand
{
public:
  QRand (int seed = -1) {
    if (seed < 0) {
      QTime midnight( 0, 0, 0 );
      seed = midnight.secsTo(QTime::currentTime());
    }
    srand(seed);
  }
  
  // return random number >= min and < max
  int getRand (int min = 0, int max = RAND_MAX) {
    return min + rand() % max; // bad random, but doesn't matter for pimping a petri net
  }
};

// Pre: dp
PrefDialog::PrefDialog(DrawProps *dp, QWidget * parent /* = 0*/, Qt::WFlags f /* = 0*/) :
  QDialog(parent, f), drawProps(dp) {
  // Setup UI
  ui.setupUi(this);
  
  // Load preferences from drawProps
  loadPrefs();
  
  // Connect some signals to prefChanged so the apply button is enabled
  connect(ui.antiAliasCheckBox, SIGNAL(clicked()), this, SLOT(prefChanged()));
  connect(ui.showIdTransitionCheckBox, SIGNAL(clicked()), this, SLOT(prefChanged()));
  connect(ui.showIdPlaceCheckBox, SIGNAL(clicked()), this, SLOT(prefChanged()));
  connect(ui.hideRoleLineEdit, SIGNAL(textChanged(const QString &)), this, SLOT(prefChanged()));
  connect(ui.transparencyCheckBox, SIGNAL(clicked()), this, SLOT(prefChanged()));
  connect(ui.colorButton, SIGNAL(changed()), this, SLOT(prefChanged()));
  
  ui.applyButton->setEnabled(false);
}

PrefDialog::~PrefDialog() {
}

void PrefDialog::loadPrefs() {
  ui.colorListWidget->clear();
  
  foreach (QString name, drawProps->brushNames()) {
    QListWidgetItem *item = new QListWidgetItem(name);
    item->setData(Qt::UserRole, drawProps->brush(name));
    ui.colorListWidget->addItem(item);
  }
  foreach (QString name, drawProps->fontNames()) {
    QListWidgetItem *item = new QListWidgetItem(name);
    item->setData(Qt::UserRole, drawProps->font(name));
    ui.fontListWidget->addItem(item);
  }
  
  ui.antiAliasCheckBox->setChecked(drawProps->option("AntiAlias"));
  ui.showIdTransitionCheckBox->setChecked(drawProps->option("ShowIdsTransitions"));
  ui.showIdPlaceCheckBox->setChecked(drawProps->option("ShowIdsPlaces"));
  ui.hideRoleLineEdit->setText(drawProps->string("HiddenRoles"));
  ui.transparencyCheckBox->setChecked(drawProps->option("ExportTransparency"));
  
  ui.colorListWidget->clearSelection();
  ui.colorButton->setDisabled(true);
  
  ui.fontListWidget->clearSelection();
  ui.changeFontButton->setDisabled(true);
}

void PrefDialog::savePrefs() {
  drawProps->setOption("AntiAlias", ui.antiAliasCheckBox->isChecked());
  drawProps->setOption("ShowIdsTransitions", ui.showIdTransitionCheckBox->isChecked());
  drawProps->setOption("ShowIdsPlaces", ui.showIdPlaceCheckBox->isChecked());
  drawProps->setString("HiddenRoles", ui.hideRoleLineEdit->text());
  drawProps->setOption("ExportTransparency", ui.transparencyCheckBox->isChecked());
  
  // Save colors (brushes)
  on_colorListWidget_currentItemChanged(NULL, ui.colorListWidget->currentItem());
  for (int i = 0; i < ui.colorListWidget->count(); i++) {
    QListWidgetItem *item = ui.colorListWidget->item(i);
    drawProps->setBrush(item->text(), item->data(Qt::UserRole).value<QBrush>());
  }
  
  // Save fonts
  for (int i = 0; i < ui.fontListWidget->count(); i++) {
    QListWidgetItem *item = ui.fontListWidget->item(i);
    drawProps->setFont(item->text(), item->data(Qt::UserRole).value<QFont>());
  }
}

void PrefDialog::on_okButton_clicked() {
  savePrefs();
  accept();
}

void PrefDialog::on_applyButton_clicked() {
  savePrefs();
  ui.applyButton->setEnabled(false);
  emit applied();
}

void PrefDialog::on_cancelButton_clicked() {
  reject();
}

void PrefDialog::on_colorListWidget_currentItemChanged(QListWidgetItem *item, QListWidgetItem *previous) {
  if (previous) {
    previous->setData(Qt::UserRole, ui.colorButton->brush());
  }
  ui.colorButton->setEnabled(item);
  if (item) {
    ui.colorButton->setBrush(item->data(Qt::UserRole).value<QBrush>());
  }
}

void PrefDialog::on_fontListWidget_currentRowChanged(int row) {
  ui.changeFontButton->setEnabled(row >= 0);
}

void PrefDialog::on_changeFontButton_clicked() {
  bool ok;
  QFont font = QFontDialog::getFont(&ok,
      ui.fontListWidget->currentItem()->data(Qt::UserRole).value<QFont>(), this);
  if (ok) {
    ui.fontListWidget->currentItem()->setData(Qt::UserRole, font);
    prefChanged();
  }
}

void PrefDialog::on_pimpButton_clicked() {
  ui.colorListWidget->clearSelection();
  QRand r;
  for (int i = 0; i < ui.colorListWidget->count(); i++) {
    ui.colorListWidget->item(i)->setData(Qt::UserRole, QBrush(QColor(r.getRand(0, 256),r.getRand(0, 256),r.getRand(0, 256))));
  }
  
  prefChanged();
}

void PrefDialog::on_resetButton_clicked() {
  ui.colorListWidget->clearSelection();
  
  DrawProps *dp = drawProps;
  drawProps = new DrawProps;
  drawProps->resetAll();
  loadPrefs();
  delete drawProps;
  drawProps = dp;
  
  prefChanged();
}

void PrefDialog::prefChanged() {
  ui.applyButton->setEnabled(true);
}
