/*
 * FILE:    transition.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "transition.h"
#include "labelplacer.h"
#include "drawprops.h"

#include <math.h>
#include "defs.h"
#include <QPainter>

#define SIGN(x) ((x)<0?-1:1)

Transition::Transition() : Connectable() {
  isEmitter = false;
  isCollector = false;
  conType = Connectable::Transition;
  type = And;
  timeMean = 0;
  timeDeviation = 0;
  fixedCost = 0;
  variableCost = 0;
  
  
  // Default size (if none is given in input file)
  size = QSize(32, 32);
}

// Pre: p && drawProps
void Transition::draw (QPainter *p, DrawProps *drawProps) {
//  qDebug("Transition::draw() id=%s", qPrintable(id));
  
  // Make rectangle
  QRectF rectF(-.5 * size.width(), // Simple rectangle to start with
               -.5 * size.height(),
               size.width(),
               size.height());
  rectF.translate(position);
  
  
  // Pen used to draw everything
  QPen pen(drawProps->pen("Default"));
  
  if (type == InlinePage) {
    // Draw rounded rectangular frame
    p->setPen(drawProps->pen("InlinePage"));
    p->setBrush(drawProps->brush("InlinePage"));
    if (rectF.width() > 0.0 && rectF.height() > 0.0) {
      p->drawRoundRect(rectF, INLINE_ROUNDING/rectF.width(), INLINE_ROUNDING/rectF.height());
    }
  } else { // type != InlinePage
    QPolygonF rect(rectF);
    
    if (type == Xor) {
      QMatrix m;
      m.translate(position.x(), position.y());
      m.scale(1.0/sqrt(2.0), 1.0/sqrt(2.0));
      m.rotate(45);
      m.translate(-position.x(), -position.y());
      rect = rect * m; // Transformed rectangle (in final position, scale and rotation)
    }
    
    // Draw shadow
    p->setPen(Qt::NoPen);
    p->setBrush(drawProps->brush("Shadow"));
    QMatrix sm;
    sm.translate(SHADOW_SIZE,SHADOW_SIZE); // Translate shadow to bottom-right
    p->drawPolygon(rect * sm);

    // Draw transition
    p->setPen(pen);
    if (isCollector) {
      p->setBrush(drawProps->brush("Collector"));   // Default: yellow
    } else if (isEmitter) {
      p->setBrush(drawProps->brush("Emitter"));      // Default: yellow
    } else if (type == Xor) {
      p->setBrush(drawProps->brush("Xor"));         // Default: orange
    } else if (type == Page) {
      p->setBrush(drawProps->brush("Page"));        // Default: blue
    } else {
      p->setBrush(drawProps->brush("Transition"));  // Default: green
    }
    p->drawPolygon(rect);
    if (timeMean > 0) {
      // Draw circle indicating a timed transition
      p->setBrush(Qt::NoBrush);
      qreal dx = pen.width();
      qreal dy = dx;
      if (type == Xor) {
        dx += (1 - 1.0/sqrt(2.0)) * .5 * size.width();
        dy += (1 - 1.0/sqrt(2.0)) * .5 * size.height();
      }
      p->drawEllipse(rectF.adjusted(dx, dy, -dx, -dy));
    }

    if (isEmitter || isCollector) {
      // Draw E or C for emitter/collector
      QFont font = drawProps->font("Name");
      font.setPixelSize(size.height()*6/10);
      p->setFont(font);
      QTextOption opt;
      opt.setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
      
      // Translations are to reduce mis-alignment
      if (isEmitter) {
        p->drawText(rectF, "E", opt);
      } else if (isCollector) {
        p->drawText(rectF, "C", opt);
      }
    }
    if (type == Page) {
      p->setPen(pen);
      // Draw magnifying-glass
      p->drawEllipse(QRectF(
                     rectF.x() + .25*size.width(),
                     rectF.y() + .25*size.height(),
                     1.0*size.width()/3,
                     1.0*size.height()/3));
      QPen thickPen(pen);
      thickPen.setWidth(2);
      p->setPen(thickPen);
      p->drawLine(QLineF(
                  rectF.x() + 1.0*size.width()*7/12 - 1.0*size.width()/6 * (1-1/sqrt(2.0)),
                  rectF.y() + 1.0*size.height()*7/12 - 1.0*size.height()/6 * (1-1/sqrt(2.0)),
                  rectF.x() + 1.0*size.width()*3/4,
                  rectF.y() + 1.0*size.height()*3/4));
    }
  }
}

// Pre: labelPlacer && drawProps
void Transition::updateLabels(LabelPlacer *labelPlacer, DrawProps *drawProps) {
  if (type == InlinePage) {
    // TODO (add 4 lines to labelPlacer)
  } else {
    labelPlacer->addRect(QRect(position.x() - size.width()/2,
                               position.y() - size.height()/2,
                               size.width(),
                               size.height()));
  }
  if (!name.isEmpty()) {
    labelPlacer->addLabel(position, name);
  } else if (drawProps->option("ShowIdsTransitions")) {
    labelPlacer->addLabel(position, id);
  }
  
  QString roleLabel;
  QStringList notRoles = drawProps->string("HiddenRoles").split(',');
  bool first = true;
  for (int i = 0; i < roles.count(); i++) {
    if (!notRoles.contains(roles[i], Qt::CaseInsensitive)) {
      roleLabel += (first?"":", ") + roles[i];
      first = false;
    }
  }
  labelPlacer->addRoleLabel(position, roleLabel);
}

const QRect Transition::getBoundingRect() const {
  QRect rect(position.x() - size.width()/2,
             position.y() - size.height()/2,
             size.width() + SHADOW_SIZE,
             size.height() + SHADOW_SIZE);
  return rect;
}

void Transition::setEmitter(bool value) {
  isEmitter = value;
}

void Transition::setCollector(bool value) {
  isCollector = value;
}

const QPointF Transition::getConnectionPoint(QPointF source) const {
  QPoint p(source.toPoint() - position);
  
  if (p.x() == 0 && p.y() == 0) {
    // Exceptional situation
    p.setX(0);
    p.setY(1);
  } else {
    // Calculate point that is most nearby p, using only coordinates (x, y),
    // with x = 0 or 1 and y = 0 or 1
    if (abs(p.x()) > abs(p.y())) {
      p.setY(static_cast<int>(1.0 * p.y() / abs(p.x())));
      p.setX(SIGN(p.x()));
    } else {
      p.setX(static_cast<int>(1.0 * p.x() / abs(p.y())));
      p.setY(SIGN(p.y()));
    }
  }
  
  QPointF q(p);
  // Scale point to match corner of rectangle
  if (type == Xor && p.x() != 0 && p.y() != 0) {
    // Correction for smaller symbol of XOR transition if
    // connencted to top-left, top-right, bottom-left or bottom-right
    //TODO: size.height()==size.width() is asserted for the next line...
    q *= .25 * size.height();
  } else {
    //TODO: size.height()==size.width() is asserted for the next line...
    q *= .5 * size.width();
  }
  
  return q + position; // Return connection point
}

const QStringList Transition::getRoles() const {
  return roles;
}

void Transition::addRole(QString value) {
  if (!value.isEmpty()) {
    roles << value;
  }
}

const QString Transition::getToolTip() const {
  QString text = Connectable::getToolTip();
  text += markup(preQuery, "Pre-query:", true);
  text += markup(postQuery, "Post-query:", true);
  text += markup(roles.join("\n"), "Performed by roles:");
  if (timeMean || timeDeviation) {
    QString time("Mean: %1\nDeviation: %2");
    time = time.arg(timeMean).arg(timeDeviation);
    text += markup(time, "Processing time:");
  }
  if (fixedCost || variableCost) {
    QString time("Fixed: %1\nPer time unit: %2");
    time = time.arg(fixedCost).arg(variableCost);
    text += markup(time, "Processing cost:");
  }
  return text;
}

bool Transition::containsPos(QPoint pos) const {
  if (type != InlinePage) {
    return QRect(position.x() - size.width()/2,
                 position.y() - size.height()/2,
                 size.width(),
                 size.height()).contains(pos);
  } else {
    QRect rect(position.x() - size.width()/2,
               position.y() - size.height()/2,
               size.width(),
               size.height());
    // Check if pos is near one of the sides of the inline page
    return  rect.adjusted(-2, -2, 2, 2).contains(pos) &&
           !rect.adjusted(2, 2, -2, -2).contains(pos);
  }
}
